
#include "mods.h"





BoundsTree::DeleteTree(Leaf *l)
	{
	Leaf *tlist[4];
	for (int i = 0; i < 4; i++)
		{
		tlist[i] = l->GetQuad(i);
		}
	delete l;
	l = NULL;

	for (i = 0; i < 4; i++)
		{
		if (tlist[i])
			DeleteTree(tlist[i]);
		}
	}

void BoundsTree::LayoutTree(Leaf *l, int depth)
	{
	depth--;
	if (depth == 0) 
		{
		l->SetBottom(TRUE);
		return;
		}

	
	for (int i =0; i < 4; i++)
		{
		l->SetQuad(i, new Leaf(l->GetBounds(i).min,l->GetBounds(i).max));
		}

	LayoutTree(l->GetQuad(0),depth);
	LayoutTree(l->GetQuad(1),depth);
	LayoutTree(l->GetQuad(2),depth);
	LayoutTree(l->GetQuad(3),depth);

	}



BoundsTree::RecurseTree(Leaf *l, int index, Box2D box)

{
	if (l->IsBottom())
		{
		BoundsList temp;
//compute normal, if facing away do not add it
		Point3 norm;
		
		Point3 a,b,c;
		Point3 vecAB, vecAC;

		a = msh.verts[msh.faces[index].v[0]];
		b = msh.verts[msh.faces[index].v[1]];
		c = msh.verts[msh.faces[index].v[2]];

		vecAB = Normalize(b-a);
		vecAC = Normalize(c-a);
		norm = CrossProd(vecAB,vecAC);

//fix set a direction here	
		if (invertNormals)
			{
			temp.faceIndex = index;
			l->faceIndex.Append(1,&temp);
			}
		else
			{
			if (norm.z > 0.0f)
				{
				temp.faceIndex = index;
				l->faceIndex.Append(1,&temp);
				}
			}
		}
	else
		{
		if ((l->GetQuad(0)) && (l->InQuad(0,box))) 
			RecurseTree(l->GetQuad(0),index,box);
		if ((l->GetQuad(1)) && (l->InQuad(1,box))) 
			RecurseTree(l->GetQuad(1),index,box);
		if ((l->GetQuad(2)) && (l->InQuad(2,box))) 
			RecurseTree(l->GetQuad(2),index,box);
		if ((l->GetQuad(3)) && (l->InQuad(3,box))) 
			RecurseTree(l->GetQuad(3),index,box);

		}
}

BoundsTree::AddFace(int index, Box2D box)
	{
	if (head)
		{
		if ((head->GetQuad(0)) && (head->InQuad(0,box))) 
			RecurseTree(head->GetQuad(0),index,box);
		if ((head->GetQuad(1)) && (head->InQuad(1,box))) 
			RecurseTree(head->GetQuad(1),index,box);
		if ((head->GetQuad(2)) && (head->InQuad(2,box))) 
			RecurseTree(head->GetQuad(2),index,box);
		if ((head->GetQuad(3)) && (head->InQuad(3,box))) 
			RecurseTree(head->GetQuad(3),index,box);
		}	
	}


BoundsTree::UpdateTreeFromBaked()
	{
	Matrix3 tm(1);
	BuildTree(&initMsh,tm);
	}

BoundsTree::BuildTree(Mesh *m, Matrix3 tm)
	{

	if (head) DeleteTree(head);

	msh = *m;

	invTm = Inverse(tm);

	for (int i =0; i < msh.numVerts; i++)
		{
		msh.verts[i] = msh.verts[i]*tm;
		}

	BuildNormals();

	//layout tree based on depth
	Point2 min;
	min.x = 0.0f;
	min.y = 0.0f;
	Point2 max;
	max.x = 1.0f;
	max.y = 1.0f;

	msh.buildBoundingBox();
	Box3 bb = msh.getBoundingBox();

	min.x = bb.pmin.x;
	min.y = bb.pmin.y;

	max.x = bb.pmax.x;
	max.y = bb.pmax.y;

	head = new Leaf(min,max);
	LayoutTree(head, depth);

//now lets scale it down to 0 to 1 range
//center the object
	Point3 center = (bb.pmin+bb.pmax) * 0.5f;
	if (normalizeMesh)
		{
		for (int i =0; i < msh.numVerts; i++)
			{
			msh.verts[i] -= bb.pmin;
//DebugPrint(" %d  point %f %f %f\n",i, msh.verts[i].x,msh.verts[i].y,msh.verts[i].z);
			}
		}
//now scale down 
	float w = 1.0f/(bb.pmax.x - bb.pmin.x);
	float h = 1.0f/(bb.pmax.y - bb.pmin.y);

	if (!normalizeMesh)
		{
		w = 1.0f;
		h = 1.0f;
		}

	w *= uscale;
	h *= vscale;


	Point3 scaleFactor(w,h,1.0f);

	for (i =0; i < msh.numVerts; i++)
		{
		msh.verts[i] *= scaleFactor;
		msh.verts[i].x += uoffset;
		msh.verts[i].y += voffset;
		}


/*
	msh.buildBoundingBox();
	Box3 newBB = msh.getBoundingBox();
	float d = (newBB.pmax.z - bb.pmin.z);

	d = 1.0f/d;
	for (i =0; i < msh.numVerts; i++)
		{
		msh.verts[i].z *= d;
		}

*/
//now stuff the points into the boudningbox list
	boundingBoxList.SetCount(msh.numFaces);
	for (i =0; i < msh.numFaces; i++)
		{
		boundingBoxList[i].SetEmpty();
		for (int j = 0; j < 3; j++)
			{
			int vid = msh.faces[i].v[j];
			Point3 p = msh.verts[vid];
			Point2 p2;
			p2.x = p.x;
			p2.y = p.y;
//DebugPrint("bound %d  point %f %f\n",i, p.x,p.y);
			boundingBoxList[i] += p2;

			}
		}
//now go through our tree adding the bounding list 
	for (i =0; i < msh.numFaces; i++)
		{
		AddFace(i,boundingBoxList[i]);
		}

	}

void BoundsTree::BuildNormals()
{

	Mesh *mesh;
	mesh = &msh;
	Face *face;	
	Point3 *verts;
	Point3 v0, v1, v2;


	face = mesh->faces;	
	verts = mesh->verts;
	vnorms.SetCount(mesh->getNumVerts());
	fnorms.SetCount(mesh->getNumFaces());

	// Compute face and vertex surface normals
	for (int i = 0; i < mesh->getNumVerts(); i++) 
		{
		vnorms[i] = VNormal();
		}
	for (i = 0; i < mesh->getNumFaces(); i++, face++) 
		{

// Calculate the surface normal
		v0 = verts[face->v[0]];
		v1 = verts[face->v[1]];
		v2 = verts[face->v[2]];
		fnorms[i] = (v1-v0)^(v2-v1);
		for (int j=0; j<3; j++) 
			{		
			vnorms[face->v[j]].AddNormal(fnorms[i],face->smGroup);
			}
		fnorms[i] = Normalize(fnorms[i]);
		}
	for (i=0; i < mesh->getNumVerts(); i++) 
		{
		vnorms[i].Normalize();
		}
}

#define EPSILON 0.000001f


BOOL BoundsTree::IsInsideMesh(Point3 hitPoint)
{

Leaf *l = head;
int ct = 0;
BOOL hit = FALSE;
if (l == NULL) 
	{
	return FALSE;
	}

while ( (l!=NULL) && (l->IsBottom() == FALSE))
	{
	int id = l->InWhichQuad(hitPoint);
	l = l->GetQuad(id);
	ct++;
	}

ct = 0;

if (l)
	{
	if (l->faceIndex.Count() == 0) 
		return FALSE;
	else
		{
		for (int i = 0; i < l->faceIndex.Count(); i++)
			{
			int faceIndex = l->faceIndex[i].faceIndex;
			Box2D b = boundingBoxList[faceIndex];
			if ( (hitPoint.x >= b.min.x) && (hitPoint.x <= b.max.x) &&
				 (hitPoint.y >= b.min.y) && (hitPoint.y <= b.max.y) )
				{
//now check bary coords
				Point3 a,b,c;
				a = msh.verts[msh.faces[faceIndex].v[0]];
				b = msh.verts[msh.faces[faceIndex].v[1]];
				c = msh.verts[msh.faces[faceIndex].v[2]];
				Point3 az,bz,cz,hitPointZ;
				az = a;
				bz = b;
				cz = c;
				az.z = 0.0f;
				bz.z = 0.0f;
				cz.z = 0.0f;
				hitPointZ = hitPoint;
				hitPointZ.z = 0.0f;
	
				Point3 bry;
				bry = BaryCoords(az, bz, cz, hitPointZ);
//if inside bary find the the z point			
				if (!( (bry.x<0.0f || bry.x>1.0f || bry.y<0.0f || bry.y>1.0f || bry.z<0.0f || bry.z>1.0f) ||
					(fabs(bry.x + bry.y + bry.z - 1.0f) > EPSILON) ))
					{
					float tz = a.z * bry.x + b.z * bry.y + c.z * bry.z; 
					if  (tz < hitPoint.z )  ct++;
					
					}
				}
			}
		}	
	}
if ((ct%2) == 1) return TRUE;
else return FALSE;

}


BOOL BoundsTree::HitMesh(Point3 hitPoint, float &z, DWORD &findex, Point3 &norm)
{

Leaf *l = head;
BOOL hit = FALSE;
if (l == NULL) 
	{
	z = 0.0f;
	return FALSE;
	}
int ct = 0;
while ( (l!=NULL) && (l->IsBottom() == FALSE))
	{
	int id = l->InWhichQuad(hitPoint);
	l = l->GetQuad(id);
	ct++;
	}
if (l)
	{
	if (l->faceIndex.Count() == 0) 
		return FALSE;
	else
		{
		for (int i = 0; i < l->faceIndex.Count(); i++)
			{
			int faceIndex = l->faceIndex[i].faceIndex;
			Box2D b = boundingBoxList[faceIndex];
			if ( (hitPoint.x >= b.min.x) && (hitPoint.x <= b.max.x) &&
				 (hitPoint.y >= b.min.y) && (hitPoint.y <= b.max.y) )
				{
//now check bary coords
				Point3 a,b,c;
				a = msh.verts[msh.faces[faceIndex].v[0]];
				b = msh.verts[msh.faces[faceIndex].v[1]];
				c = msh.verts[msh.faces[faceIndex].v[2]];
				Point3 az,bz,cz,hitPointZ;
				az = a;
				bz = b;
				cz = c;
				az.z = 0.0f;
				bz.z = 0.0f;
				cz.z = 0.0f;
				hitPointZ = hitPoint;
				hitPointZ.z = 0.0f;
	
				Point3 bry;
				bry = BaryCoords(az, bz, cz, hitPointZ);
//if inside bary find the the z point			
				if (!( (bry.x<0.0f || bry.x>1.0f || bry.y<0.0f || bry.y>1.0f || bry.z<0.0f || bry.z>1.0f) ||
					(fabs(bry.x + bry.y + bry.z - 1.0f) > EPSILON) ))
					{
					float tz = a.z * bry.x + b.z * bry.y + c.z * bry.z; 
					if ( (tz > z ) || (hit == FALSE) )  
						{
						z = tz;
						findex = faceIndex;
						norm = fnorms[faceIndex];
						}
					hit = TRUE;
					}
				}
			}
		}	
	}

return hit;
}




BOOL BoundsTree::HitMesh(Point3 hitPoint, Point3 &norm)
{
norm = Point3 (0.0f,0.0f,0.0f);
Leaf *l = head;
BOOL hit = FALSE;
int nia,nib,nic;
Point3 ibary;
DWORD smgroup;
int ifaceID;

float z = 0.0f;

if (l == NULL) 
	{
	
	return FALSE;
	}
int ct = 0;
while ( (l!=NULL) && (l->IsBottom() == FALSE))
	{
	int id = l->InWhichQuad(hitPoint);
	l = l->GetQuad(id);
	ct++;
	}
if (l)
	{
	if (l->faceIndex.Count() == 0) 
		return FALSE;
	else
		{
		for (int i = 0; i < l->faceIndex.Count(); i++)
			{
			int faceIndex = l->faceIndex[i].faceIndex;
			Box2D b = boundingBoxList[faceIndex];
			if ( (hitPoint.x >= b.min.x) && (hitPoint.x <= b.max.x) &&
				 (hitPoint.y >= b.min.y) && (hitPoint.y <= b.max.y) )
				{
//now check bary coords
				int ia,ib,ic;
				
				ia = msh.faces[faceIndex].v[0];
				ib = msh.faces[faceIndex].v[1];
				ic = msh.faces[faceIndex].v[2];
				Point3 a,b,c;
				a = msh.verts[ia];
				b = msh.verts[ib];
				c = msh.verts[ic];
				Point3 az,bz,cz,hitPointZ;
				az = a;
				bz = b;
				cz = c;
				az.z = 0.0f;
				bz.z = 0.0f;
				cz.z = 0.0f;
				hitPointZ = hitPoint;
				hitPointZ.z = 0.0f;
	
				Point3 bry;
				bry = BaryCoords(az, bz, cz, hitPointZ);
//if inside bary find the the z point			
				if (!( (bry.x<0.0f || bry.x>1.0f || bry.y<0.0f || bry.y>1.0f || bry.z<0.0f || bry.z>1.0f) ||
					(fabs(bry.x + bry.y + bry.z - 1.0f) > EPSILON) ))
					{
					float tz = a.z * bry.x + b.z * bry.y + c.z * bry.z; 
					if ( (tz > z ) || (hit == FALSE) )  
						{
						z = tz;
						hit = TRUE;
						nia = ia;
						nib = ib;
						nic = ic;
						ifaceID = faceIndex;
						smgroup = msh.faces[faceIndex].getSmGroup();
						ibary = bry;
						}
					}
				}
			}
		}	
	}
if (hit)
	{
	Point3 normA, normB, normC;
	normA = vnorms[nia].GetNormal(smgroup);
	normB = vnorms[nib].GetNormal(smgroup);
	normC = vnorms[nic].GetNormal(smgroup);
	norm = (normA * ibary.x) + (normB * ibary.y) + (normC * ibary.z);
//	norm = fnorms[ifaceID];
	}

return hit;
}


BoundsTree::SetInitMesh(Mesh *m)
	{
	initMsh = *m;
	}
Mesh BoundsTree::GetInitMesh()
	{
	return initMsh;
	}

IOResult BoundsTree::Load(ILoad *iload)
	{
	ULONG nb;
	IOResult res;
	int id;
	int fct = 0;
	int vct = 0;
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(id=iload->CurChunkID())  {
			case MESH_VERTEXCOUNT_CHUNK:
				iload->Read(&vct,sizeof(vct),&nb);
				initMsh.setNumVerts(vct);
				break;
			case MESH_FACECOUNT_CHUNK:
				iload->Read(&fct,sizeof(fct),&nb);
				initMsh.setNumFaces(fct);
				break;
			case MESH_VERTEXDATA_CHUNK:
				iload->Read(initMsh.verts,sizeof(Point3)*vct,&nb);
				
				break;
			case MESH_FACEDATA_CHUNK:
				iload->Read(initMsh.faces,sizeof(Face)*fct,&nb);
				break;
			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}


	return IO_OK;
	
	}

IOResult BoundsTree::Save(ISave *isave) 
	{
	ULONG nb;
	// Save common stuff

	int vct,fct;
	vct = initMsh.numVerts;
	fct = initMsh.numFaces;
	//save vertex count
	isave->BeginChunk(MESH_VERTEXCOUNT_CHUNK);
	isave->Write(&vct,sizeof(vct),&nb);
	isave->EndChunk();

	//save face count
	isave->BeginChunk(MESH_FACECOUNT_CHUNK);
	isave->Write(&fct,sizeof(fct),&nb);
	isave->EndChunk();

	//save vertex data
	isave->BeginChunk(MESH_VERTEXDATA_CHUNK);
	isave->Write(initMsh.verts,sizeof(Point3)*vct,&nb);
	isave->EndChunk();

	//save face data
	isave->BeginChunk(MESH_FACEDATA_CHUNK);
	isave->Write(initMsh.faces,sizeof(Face)*fct,&nb);
	isave->EndChunk();

	return IO_OK;

	}
