/**********************************************************************
 *<
	FILE: mods.h

	DESCRIPTION:

	CREATED BY: Rolf Berteig (based on prim.h)

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#ifndef __MODS__H
#define __MODS__H

#include "Max.h"
//#include "reslib.h"
#include "modsres.h"
#include "iparamm.h"

#define PBLOCK_REF	0

#define LAZYID 0xDE17A346, 0x8A41E2Ab
#include "msgpack.h"
#include <wintab.h>

#define PACKETDATA	( PK_BUTTONS | PK_NORMAL_PRESSURE)
#define PACKETMODE	PK_BUTTONS

#include <pktdef.h>



TCHAR *GetString(int id);

#define EDITMESH_CLASS_ID			0x00050
#define EDITSPLINE_CLASS_ID			0x00060
#define EDITPATCH_CLASS_ID			0x00070
#define EDITLOFT_CLASS_ID			0x00080

#define CLUSTOSM_CLASS_ID			0x25215824

#define RESET_XFORM_CLASS_ID		0x8d562b81
#define CLUSTNODEOSM_CLASS_ID		0xc4d33

extern ClassDesc* GetAFRModDesc();
// in mods.cpp
extern HINSTANCE hInstance;

// For 'Supports Object of Type' rollups
extern BOOL CALLBACK DefaultSOTProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);

#define BIGFLOAT	float(999999)

#define NEWSWMCAT	_T("Modifiers")

#define USETABLET

#define STARTSTROKE		1
#define INSTROKE		2
#define ENDSTROKE		3
#define OUTOFSTROKE		4


#define MESH_CHUNK 0x1010

#define MESH_VERTEXCOUNT_CHUNK 0x1020
#define MESH_FACECOUNT_CHUNK 0x1030

#define MESH_VERTEXDATA_CHUNK 0x1040
#define MESH_FACEDATA_CHUNK 0x1050


class VNormal {
	public:
		Point3 norm;
		DWORD smooth;
		VNormal *next;
		BOOL init;

		VNormal() {smooth=0;next=NULL;init=FALSE;norm=Point3(0,0,0);}
		VNormal(Point3 &n,DWORD s) {next=NULL;init=TRUE;norm=n;smooth=s;}
		~VNormal() {delete next;}
		void AddNormal(Point3 &n,DWORD s);
		Point3 &GetNormal(DWORD s);
		void Normalize();
	};




class Leaf;

class BoundsList
{
public:
	int faceIndex;
	BoundsList()
		{
		faceIndex = -1;
		}
};

class Leaf
{

private:
	Box2D bounds[4];
	BOOL bottom;
	
public:

	Leaf *quad[4];
	Tab<BoundsList> faceIndex;

	Leaf(Point2 min, Point2 max)
		{
		bottom = FALSE;
		quad[0] = NULL;
		quad[1] = NULL;
		quad[2] = NULL;
		quad[3] = NULL;
		Point2 mid = (min+max)*0.5f;

		bounds[0].min = min;
		bounds[0].max = mid;

		bounds[1].min.x = mid.x;
		bounds[1].min.y = min.y;
		bounds[1].max.x = max.x;
		bounds[1].max.y = mid.y;

		bounds[2].min.x = min.x;
		bounds[2].min.y = mid.y;
		bounds[2].max.x = mid.x;
		bounds[2].max.y = max.y;

		bounds[3].min = mid;
		bounds[3].max = max;

		}
	BOOL InQuad(int i, Box2D box)
		{
		Box2D b = bounds[i];
		if (box.max.x < b.min.x) return FALSE;
		if (box.min.x > b.max.x) return FALSE;

		if (box.max.y < b.min.y) return FALSE;
		if (box.min.y > b.max.y) return FALSE;

		return TRUE;

		}
	int InWhichQuad(Point3 p)
		{
		if ( (p.x >=bounds[0].min.x) && (p.x < bounds[0].max.x) &&
			 (p.y >=bounds[0].min.y) && (p.y < bounds[0].max.y) )
			 return 0;
		else if ( (p.x >=bounds[1].min.x) && (p.x < bounds[1].max.x) &&
			 (p.y >=bounds[1].min.y) && (p.y < bounds[1].max.y) )
			 return 1;
		else if ( (p.x >=bounds[2].min.x) && (p.x < bounds[2].max.x) &&
			 (p.y >=bounds[2].min.y) && (p.y < bounds[2].max.y) )
			 return 2;
		else if ( (p.x >=bounds[3].min.x) && (p.x < bounds[3].max.x) &&
			 (p.y >=bounds[3].min.y) && (p.y < bounds[3].max.y) )
			 return 3;
		return -1;
		}

	Box2D GetBounds(int i) {	if ( (i>=0) && (i<4) ) 
								return bounds[i];
								else return Box2D();}
	
	Leaf *GetQuad(int i) {	if ( (i>=0) && (i<4) ) 
							return quad[i];
							else return NULL;
						}
	void SetQuad(int i, Leaf *l) { if ( (i>=0) && (i<4) ) quad[i] = l;}

	BOOL IsBottom() {return bottom;}
	void SetBottom (BOOL b) { bottom = b;}

};

class BoundsTree
{


	BOOL invertNormals;
	BOOL normalizeMesh;
	int depth;

	float uoffset, voffset, uscale, vscale;

	Tab<VNormal> vnorms;
	Tab<Point3> fnorms;
	

	

	Mesh msh;
	Mesh initMsh;

	void LayoutTree(Leaf *l, int depth);

	RecurseTree(Leaf *l, int index, Box2D box);
	AddFace(int index, Box2D box);
	Tab<Box2D> boundingBoxList;

	void BuildNormals();


public:


	Leaf *head;

	Matrix3 invTm;

	BoundsTree()
		{
		normalizeMesh = TRUE;
		uoffset = 0.0f;
		voffset = 0.0f;
		uscale = 1.0f;
		vscale = 1.0f;
		depth = 6;
		head = NULL;
		}
	DeleteTree(Leaf *l);
	UpdateTreeFromBaked();
	BuildTree(Mesh *m, Matrix3 tm);

	SetInitMesh(Mesh *m);
	Mesh GetInitMesh();
	
	IOResult Load(ILoad *iload);
	IOResult Save(ISave *isave); 

	void CloneFrom(BoundsTree tree)
		{
		initMsh = tree.initMsh;
		}

	BOOL IsInsideMesh(Point3 hitPoint);
	BOOL HitMesh(Point3 hitPoint, float &z, DWORD &findex, Point3 &norm);
	BOOL HitMesh(Point3 hitPoint, Point3 &norm);


	void InvertNorms(BOOL n) { invertNormals = n;}

	void SetDepth(int d) { depth = d;}
	void SetNormalizeMesh(int n) { normalizeMesh = n;}
	void SetPlacement(float ou, float ov, float su, float sv) 
			{ 
			uoffset = ou;
			voffset = ov;
			uscale = su;
			vscale = sv;
			}

private:
};



class NullView: public View {
	public:
		Point2 ViewToScreen(Point3 p) { return Point2(p.x,p.y); }
		NullView() { worldToView.IdentityMatrix(); screenW=640.0f; screenH = 480.0f; }
	};


class PaintMode;

class AFRMod : public Modifier {	
	public:
	
#ifdef USETABLET
//tablet stuff

		int strokeState;
		UINT pressure;
		float fpressure;


#endif

		BoundsTree meshSearchTree;

		BOOL isMesh;
		BOOL redoNormalList;
		Tab<VNormal> vnorms;
		Tab<Point3> fnorms;
		void BuildNormals(Mesh *mesh);
		Mesh msh;



		IParamBlock *pblock;
		static PaintMode *paintMode;

		Tab<Point3> DeltaList;
		Tab<Point3> TempList;
		Tab<BOOL> SelectVertList;
		int SelectMode;

//		Control *p1;
		BYTE sel[1];

		static IObjParam *ip;
		static IParamMap *pmapParam;
		
		AFRMod();
		~AFRMod() {
				DeleteAllRefsFromMe();
				pblock = NULL;
//				p1 = NULL;
				}

static	float strength, eradius;
		float strengthsquared,eradiussquared;
static	float hardness;
static	float effect;
		int Addative;
		IPoint2 pp;

		HWND inkHWND;
static		int inkType;


		// From Animatable
		void DeleteThis() { delete this; }
		void GetClassName(TSTR& s) {s = GetString(IDS_RB_AFRMOD);}  
		virtual Class_ID ClassID() { return Class_ID(LAZYID);}		
		void BeginEditParams(IObjParam  *ip, ULONG flags,Animatable *prev);
		void EndEditParams(IObjParam *ip,ULONG flags,Animatable *next);		
		RefTargetHandle Clone(RemapDir& remap = NoRemap());
		TCHAR *GetObjectName() {return GetString(IDS_RB_AFRMOD);}

		// From modifier
		ChannelMask ChannelsUsed()  {return PART_GEOM|PART_TOPO|PART_SELECT|PART_SUBSEL_TYPE;}
		ChannelMask ChannelsChanged() {return PART_GEOM;}
		Class_ID InputType() {return defObjectClassID;}
		void ModifyObject(TimeValue t, ModContext &mc, ObjectState *os, INode *node);
		Interval LocalValidity(TimeValue t);

		// From BaseObject
		CreateMouseCallBack* GetCreateMouseCallBack() {return NULL;} 

		int NumRefs() {return 1;}
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

		int NumSubs() {return 1;}
		Animatable* SubAnim(int i){return pblock;}; 
		TSTR SubAnimName( int i  ) {return TSTR(GetString(IDS_RB_PARAMETERS));};

		RefResult NotifyRefChanged( Interval changeInt,RefTargetHandle hTarget, 
		   PartID& partID, RefMessage message);



		void EnterNormalAlign();
		void ExitNormalAlign();

		IOResult Load(ILoad *iload);
		IOResult Save(ISave *isave);

		int Display(TimeValue t, INode* inode, ViewExp *vpt, int flagst, ModContext *mc);
		void DrawCrossSection(Point3 a, Point3 align, float length,float strenght, GraphicsWindow *gw);

		BOOL inPaint;
		Point3 paintPoint, paintNorm;

	};


	class AFRUndoList
{
public:
int vertindex;
Point3 delta;
};

// Flags
#define CONTROL_FIT			(1<<0)
#define CONTROL_CENTER		(1<<1)
#define CONTROL_ASPECT		(1<<2)
#define CONTROL_UNIFORM		(1<<3)
#define CONTROL_HOLD		(1<<4)
#define CONTROL_INIT		(1<<5)
#define CONTROL_OP			(CONTROL_FIT|CONTROL_CENTER|CONTROL_ASPECT|CONTROL_UNIFORM)
#define CONTROL_INITPARAMS	(1<<10)


#define CID_FACEALIGNMAP 	0x4f298c7a
#define CID_REGIONFIT 		0x4f298c7b


class PaintMouseProc : public MouseCallBack {
	public:
		AFRMod *mod;
		IObjParam *ip;
		PaintMouseProc(AFRMod *m,IObjParam *i) {mod=m;ip=i;}
		int proc(HWND hWnd, int msg, int point, int flags, IPoint2 m);
		void PaintMap(HWND hWnd,IPoint2 m);
		void HitMap(HWND hWnd,IPoint2 m);

		void MousePointDown(int point, int flags, IPoint2 m);
		void MousePointUp(int point, int flags, IPoint2 m);
		void MouseMove();
	};

class PaintMode : public CommandMode {
	public:
		ChangeFGObject fgProc;
		PaintMouseProc proc;
		IObjParam *ip;
		AFRMod *mod;

		PaintMode(AFRMod *m,IObjParam *i) 
			: fgProc(m), proc(m,i) {ip=i;mod=m;}

		int Class() {return MOVE_COMMAND;}		
		int ID() {return CID_FACEALIGNMAP;}
		MouseCallBack *MouseProc(int *numPoints) {*numPoints=2;return &proc;}
		ChangeForegroundCallback *ChangeFGProc() {return &fgProc;}
		BOOL ChangeFG(CommandMode *oldMode) {return oldMode->ChangeFGProc() != &fgProc;}
		void EnterMode();
		void ExitMode();
	};

// A restore object to save the position of control points.
class MoveRestore : public RestoreObj {
	public:		
		AFRMod *mod;
		Tab<AFRUndoList> PListUndo;
		MoveRestore(AFRMod *m) {
			mod = m;
			PListUndo.ZeroCount();
			for (int i=0; i<mod->TempList.Count(); i++) {
				if (Length(mod->TempList[i]) > 0.000f)
					{
					AFRUndoList t;
					t.vertindex = i;
					t.delta = mod->TempList[i];
					PListUndo.Append(1,&t,1);
//					PListRedo.Append(1,&t,1);
					}
				}
			}

		void Restore(int isUndo) {
			// if we're undoing, save a redo state
/*			if (isUndo) {
				for (int i=0; i<mod->NumCtrlPts(); i++) {
					redo[i] = mod->GetPt(i);
					}
				}
*/
			if (isUndo) {
				for (int i=0; i<PListUndo.Count(); i++) {
					mod->DeltaList[PListUndo[i].vertindex] -= PListUndo[i].delta;
//					mod->GetPt(i) = undo[i];
					}
				mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
				}
			}

		void Redo() {
			for (int i=0; i<PListUndo.Count(); i++) {
				mod->DeltaList[PListUndo[i].vertindex] += PListUndo[i].delta;
//				mod->GetPt(i) = undo[i];
				}
//			for (int i=0; i<mod->NumCtrlPts(); i++) {
//				mod->GetPt(i) = redo[i];
//				}
			mod->NotifyDependents(FOREVER,PART_GEOM,REFMSG_CHANGE);
			}
//		void EndHold() {
//			mod->ClearAFlag(A_HELD);
//			}
	TSTR Description()

		{ return TSTR(_T("Paint Deform")); }
	};




#endif

