/**********************************************************************
 *<
	FILE: afregion.cpp

	DESCRIPTION:  Affect region modifier

	CREATED BY: Rolf Berteig

	HISTORY: 10/16/96

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/

#include "mods.h"







float AFRMod::strength = 2.0f;
float AFRMod::eradius = 30.0f;
float AFRMod::hardness = 1.0f;
float AFRMod::effect = 1.0f;
int AFRMod::inkType = 0;






PaintMode*      AFRMod::paintMode = NULL;


static void MatrixFromNormal(Point3& normal, Matrix3& mat)
	{
	Point3 vx;
	vx.z = .0f;
	vx.x = -normal.y;
	vx.y = normal.x;	
	if ( vx.x == .0f && vx.y == .0f ) {
		vx.x = 1.0f;
		}
	mat.SetRow(0,vx);
	mat.SetRow(1,normal^vx);
	mat.SetRow(2,normal);
	mat.SetTrans(Point3(0,0,0));
	mat.NoScale();
	}




class AFRDeformer : public Deformer {
	public:		

		AFRMod *mod;
		AFRDeformer(AFRMod *m) {mod = m;}
		Point3 Map(int i, Point3 p);
	};


//--- ClassDescriptor and class vars ---------------------------------

IParamMap       *AFRMod::pmapParam = NULL;
IObjParam       *AFRMod::ip        = NULL;
		

class AFRClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading = FALSE) { return new AFRMod; }
	const TCHAR *	ClassName() { return GetString(IDS_RB_PMOD); }
	SClass_ID		SuperClassID() { return OSM_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(LAZYID); }
	const TCHAR* 	Category() { return GetString(IDS_RB_DEFDEFORMATIONS);}
	};

static AFRClassDesc afrDesc;
extern ClassDesc* GetAFRModDesc() {return &afrDesc;}


//--- Parameter map/block descriptors -------------------------------

#define PB_STR			0
#define PB_ERADIUS		1
#define PB_HARDNESS		2
#define PB_ADDATIVE		3
#define PB_EFFECT		4

//
//
// Parameters

static ParamUIDesc descParam[] = {
	// Falloff
	ParamUIDesc(
		PB_STR,
		EDITTYPE_FLOAT,
		IDC_AFR_STR,IDC_AFR_STRSPIN,
		-BIGFLOAT,BIGFLOAT,
		SPIN_AUTOSCALE),

	// Settle
	ParamUIDesc(
		PB_ERADIUS,
		EDITTYPE_FLOAT,
		IDC_AFR_ERADIUS,IDC_AFR_ERADIUS_SPIN,
		0.0f,BIGFLOAT,
		SPIN_AUTOSCALE),

	// Waves
	ParamUIDesc(
		PB_HARDNESS,
		EDITTYPE_FLOAT,
		IDC_AFR_HARDNESS,IDC_AFR_HARDNESS_SPIN,
		0.0f,1.0f,
		SPIN_AUTOSCALE),

	// Addative
   ParamUIDesc(PB_ADDATIVE,TYPE_SINGLECHEKBOX,IDC_ADDATIVE),

	// Effect
	ParamUIDesc(
		PB_EFFECT,
		EDITTYPE_FLOAT,
		IDC_AFR_EFFECT,IDC_AFR_EFFECT_SPIN,
		-5.0f,5.0f,
		SPIN_AUTOSCALE),


	};
#define PARAMDESC_LENGTH	5

static ParamBlockDescID descVer0[] = {
	{ TYPE_FLOAT, NULL, FALSE,  0 },		// Falloff	
	{ TYPE_FLOAT, NULL, FALSE,  1 },		// Settle	
	{ TYPE_FLOAT, NULL, FALSE,  2 },		// Waves
	{ TYPE_INT, NULL, FALSE,  3 },		// Addative
	};
static ParamBlockDescID descVer1[] = {
	{ TYPE_FLOAT, NULL, FALSE,  0 },		// Falloff	
	{ TYPE_FLOAT, NULL, FALSE,  1 },		// Settle	
	{ TYPE_FLOAT, NULL, FALSE,  2 },		// Waves
	{ TYPE_INT, NULL, FALSE,  3 },		// Addative
	{ TYPE_INT, NULL, TRUE,  4 },		// Effect
	};
#define PBLOCK_LENGTH	5

#define CURRENT_VERSION	1

// Array of old versions
static ParamVersionDesc versions[] = {
	ParamVersionDesc(descVer0,3,0),
	};
#define NUM_OLDVERSIONS	1
static ParamVersionDesc curVersion(descVer1,PBLOCK_LENGTH,CURRENT_VERSION);

//---------------------------------------------------------------------
// UI stuff


class PaintRightMenu : public RightClickMenu {
	private:
		AFRMod *ep;
	public:
		void Init(RightClickMenuManager* manager, HWND hWnd, IPoint2 m);
		void Selected(UINT id);
		void SetMod(AFRMod *ep) { this->ep = ep; }
	};

void PaintRightMenu::Init(RightClickMenuManager* manager, HWND hWnd, IPoint2 m) {
		int flags1,flags2;
		flags1 = flags2 = MF_STRING;
		flags1 |= MF_CHECKED;


		manager->AddMenu(this, flags1, 0, "Move Normal");
		manager->AddMenu(this, flags2, 1, "Move XY Screen");
		manager->AddMenu(this, flags2, 2, "Move Z Screen");
		manager->AddMenu(this, flags2, 3, "Move X Local");
		manager->AddMenu(this, flags2, 4, "Move Y Local");
		manager->AddMenu(this, flags2, 5, "Move Z Local");
/*
		manager->AddMenu(this, MF_SEPARATOR, 0, NULL);
		manager->AddMenu(this, flags2, 10, "Str -32");
		manager->AddMenu(this, flags2, 11, "Str -16");
		manager->AddMenu(this, flags2, 12, "Str -8");
		manager->AddMenu(this, flags2, 13, "Str -4");
		manager->AddMenu(this, flags2, 14, "Str -2");
		manager->AddMenu(this, flags2, 15, "Str -1");
		manager->AddMenu(this, flags2, 16, "Str 1");
		manager->AddMenu(this, flags2, 17, "Str 2");
		manager->AddMenu(this, flags2, 18, "Str 4");
		manager->AddMenu(this, flags2, 19, "Str 8");
		manager->AddMenu(this, flags2, 20, "Str 16");
		manager->AddMenu(this, flags2, 21, "Str 32");
		manager->AddMenu(this, MF_SEPARATOR, 0, NULL);
		manager->AddMenu(this, flags2, 30, "Rad 1");
		manager->AddMenu(this, flags2, 31, "Rad 2");
		manager->AddMenu(this, flags2, 32, "Rad 4");
		manager->AddMenu(this, flags2, 33, "Rad 8");
		manager->AddMenu(this, flags2, 34, "Rad 16");
		manager->AddMenu(this, flags2, 35, "Rad 32");
		manager->AddMenu(this, flags2, 36, "Rad 64");
*/
		}
	

void PaintRightMenu::Selected(UINT id) {
//change ink type
	if (id<=5)
		{
		}
//chnage strength	
	else if (id<=21)
		{
		}
//change radius	
	else if (id<=36)
		{
		}
	}

PaintRightMenu pMenu;

#ifdef USETABLET

HCTX static NEAR TabletInit(HWND hWnd)
{
	LOGCONTEXT lcMine;


	UINT p,rp; 
	rp = WTInfo(WTI_DEVICES, DVC_NPRESSURE, &p );


	/* get default region */
	WTInfo(WTI_DEFCONTEXT, 0, &lcMine);

	/* modify the digitizing region */
	wsprintf(lcMine.lcName, "PrsTest Digitizing %x", hInstance);
	lcMine.lcOptions |= CXO_MESSAGES|CXO_SYSTEM;
	lcMine.lcPktData = PACKETDATA;
	lcMine.lcPktMode = PACKETMODE;
	lcMine.lcMoveMask = PACKETDATA;
	lcMine.lcBtnUpMask = lcMine.lcBtnDnMask;

	/* output in 10000 x 10000 grid */
	lcMine.lcOutOrgX = lcMine.lcOutOrgY = 0;
	lcMine.lcOutExtX = 10000;
	lcMine.lcOutExtY = 10000;

	/* open the region */
	return WTOpen(hWnd, &lcMine, TRUE);

}

#endif

//--- MapDlgProc -----------------------------------------------------

class PaintDlgProc : public ParamMapUserDlgProc {
	public:
		AFRMod *mod;		
		PaintDlgProc(AFRMod *m) {mod = m;}		
		BOOL DlgProc(TimeValue t,IParamMap *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);		
		void DeleteThis() {delete this;}
	};


BOOL PaintDlgProc::DlgProc(
		TimeValue t,IParamMap *map,HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
	{
	int id = LOWORD(wParam);
	int code = HIWORD(wParam);

#ifdef USETABLET
	static HCTX hTab = NULL;
	PACKET pkt;
	static UINT prsOld, prsNew;
#endif


	switch (msg) {
		case WM_INITDIALOG:
			SendMessage(GetDlgItem(hWnd,IDC_PAINT),CC_COMMAND,CC_CMD_SET_TYPE,CBT_CHECK);
//fillout combo box
			mod->inkHWND = GetDlgItem(hWnd, IDC_INK);
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Move Normal");
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Move XY Screen");
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Move Z Screen");
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Move XY Local");
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Move Z Local");
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Smooth");
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Tesselate");
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Select Verts");
			SendMessage(mod->inkHWND, CB_ADDSTRING, 0, (LPARAM)"Select Faces");
			SendMessage( GetDlgItem(hWnd,IDC_INK), CB_SETCURSEL, mod->inkType , 0 );
//			mod->inkType = 0;

//			pMenu.SetMod(mod);
//			mod->ip->GetRightClickMenuManager()->Register(&pMenu);

#ifdef USETABLET
			hTab = TabletInit(hWnd);
			if (!hTab) {
				MessageBox(NULL, " Could Not Open Tablet Context.", "WinTab",
							MB_OK | MB_ICONHAND);
			}
#endif


			break;




#ifdef USETABLET
		case WT_PACKET:
			if (WTPacket((HCTX)lParam, wParam, &pkt)) 
				{
				if (mod->inPaint)
					{
					if (HIWORD(pkt.pkButtons)==TBN_DOWN) 
						{
//						MessageBeep(0);
						mod->strokeState = STARTSTROKE;
						}
					else if (HIWORD(pkt.pkButtons)==TBN_UP) 
						{
//						MessageBeep(MB_ICONQUESTION);
						mod->strokeState = ENDSTROKE;
						mod->fpressure  = 1.0f;
					}

					mod->pressure = pkt.pkNormalPressure;
					mod->fpressure = (float)(mod->pressure)/1024.0f;
				
					TSTR name;
					name.printf("%1.2f (%d)",mod->fpressure,mod->pressure);
					SetWindowText(GetDlgItem(hWnd,IDC_WEIGHT),name);

					if (mod->pressure != 0) mod->strokeState = INSTROKE;
					}
				}
			
			break;
#endif

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDC_PAINT:
					if (mod->ip->GetCommandMode()->ID()==CID_FACEALIGNMAP) {
						mod->ip->SetStdCommandMode(CID_OBJMOVE);
					} else {
						mod->ip->SetCommandMode(mod->paintMode);
						}
					break;
				case IDC_INK: {
					if (code==CBN_SELCHANGE) {
						mod->inkType = SendMessage( GetDlgItem(hWnd,IDC_INK), CB_GETCURSEL, 0, 0 );
						}
					break;
					}

				}
			break;
		}
	return FALSE;
	}




//--- Affect region mod methods -------------------------------

AFRMod::AFRMod() 
	{
	MakeRefByID(
		FOREVER, PBLOCK_REF, 
		CreateParameterBlock(
			descVer0, PBLOCK_LENGTH, CURRENT_VERSION));
	
	sel[0] = 0;
	pblock->SetValue(PB_STR,0,strength);
	pblock->SetValue(PB_ERADIUS,0,eradius);
	pblock->SetValue(PB_HARDNESS,0,hardness);
	pblock->SetValue(PB_EFFECT,0,effect);
	inkType = 0;
	inPaint = FALSE;
	strokeState = OUTOFSTROKE;

	isMesh = FALSE;
	redoNormalList = TRUE;
	}

void AFRMod::EnterNormalAlign()
	{	
	SendMessage(GetDlgItem(pmapParam->GetHWnd(),IDC_PAINT),CC_COMMAND,CC_CMD_SET_STATE,1);
	}

void AFRMod::ExitNormalAlign()
	{	
	SendMessage(GetDlgItem(pmapParam->GetHWnd(),IDC_PAINT),CC_COMMAND,CC_CMD_SET_STATE,0);
	}


void AFRMod::BeginEditParams(
		IObjParam  *ip, ULONG flags,Animatable *prev)
	{
	fpressure  = 1.0f;
	this->ip = ip;
	// Add our sub object type
	TSTR type1(GetString(IDS_RB_AFRPOINTS));

	// Create sub object editing modes.



/*
	ModContextList mcList;
	INodeTab nodes;
	ip->GetModContexts(mcList,nodes);
	assert(nodes.Count());

	tnode = nodes[0];
*/

//	if (SelfNode) ReplaceReference(SELFNODE_REF,node,TRUE);
//	else MakeRefByID(FOREVER,SELFNODE_REF,node);
	


//	SelfNode = nodes[0];

//	NotifyDependents(FOREVER, 0, REFMSG_CHANGE);


	paintMode = new PaintMode(this,ip);

	
	TimeValue t = ip->GetTime();
	NotifyDependents(Interval(t,t), PART_ALL, REFMSG_BEGIN_EDIT);
	NotifyDependents(Interval(t,t), PART_ALL, REFMSG_MOD_DISPLAY_ON);
	SetAFlag(A_MOD_BEING_EDITED);

	pmapParam = CreateCPParamMap(
		descParam,PARAMDESC_LENGTH,
		pblock,
		ip,
		hInstance,
		MAKEINTRESOURCE(IDD_AFRPARAM),
		GetString(IDS_RB_PARAMETERS),
		0);	
	pmapParam->SetUserDlgProc(new PaintDlgProc(this));

	}

void AFRMod::EndEditParams(
		IObjParam *ip,ULONG flags,Animatable *next)
	{
	this->ip = NULL;
	TimeValue t = ip->GetTime();

	// NOTE: This flag must be cleared before sending the REFMSG_END_EDIT
	ClearAFlag(A_MOD_BEING_EDITED);

	NotifyDependents(Interval(t,t), PART_ALL, REFMSG_END_EDIT);
	NotifyDependents(Interval(t,t), PART_ALL, REFMSG_MOD_DISPLAY_OFF);

	ip->DeleteMode(paintMode);
	delete paintMode; paintMode = NULL;

	DestroyCPParamMap(pmapParam);


	}

RefTargetHandle AFRMod::Clone(RemapDir& remap)
	{
	AFRMod *mod = new AFRMod();
	mod->ReplaceReference(PBLOCK_REF,pblock->Clone(remap));
	BaseClone(this, mod, remap);
//	mod->ReplaceReference(SELFNODE_REF,SelfNode->Clone(remap));
//	mod->SelfNode=NULL;
	return mod;
	}


void AFRMod::BuildNormals(Mesh *mesh)
{

	Face *face;	
	Point3 *verts;
	Point3 v0, v1, v2;


	face = mesh->faces;	
	verts = mesh->verts;
	vnorms.SetCount(mesh->getNumVerts());
	fnorms.SetCount(mesh->getNumFaces());

	// Compute face and vertex surface normals
	for (int i = 0; i < mesh->getNumVerts(); i++) 
		{
		vnorms[i] = VNormal();
		}
	for (i = 0; i < mesh->getNumFaces(); i++, face++) 
		{

// Calculate the surface normal
		v0 = verts[face->v[0]];
		v1 = verts[face->v[1]];
		v2 = verts[face->v[2]];
		fnorms[i] = (v1-v0)^(v2-v1);
		for (int j=0; j<3; j++) 
			{		
			vnorms[face->v[j]].AddNormal(fnorms[i],face->smGroup);
			}
		fnorms[i] = Normalize(fnorms[i]);
		}
	for (i=0; i < mesh->getNumVerts(); i++) 
		{
		vnorms[i].Normalize();
		}
}


void AFRMod::ModifyObject(
		TimeValue t, ModContext &mc, ObjectState *os, INode *node)
	{	
	int nv = os->obj->NumPoints();

	if (DeltaList.Count() == 0)
		{
		Point3 p(0.0f,0.0f,0.0f);
		BOOL b = FALSE;
		for (int i = 0; i < nv; i++)
			{
			DeltaList.Append(1,&p,1);
			TempList.Append(1,&p,1);
			SelectVertList.Append(1,&b,1);
			}
		}


	Interval valid = FOREVER;

	pblock->GetValue(PB_STR,t,strength,valid);		
	pblock->GetValue(PB_ERADIUS,t,eradius,valid);		
	pblock->GetValue(PB_HARDNESS,t,hardness,valid);		
	pblock->GetValue(PB_ADDATIVE,t,Addative,valid);		
	pblock->GetValue(PB_EFFECT,t,effect,valid);		
	

	strengthsquared = (float) fabs(strength);

	eradiussquared = eradius * eradius;


	Interval iv = os->obj->ChannelValidity (t, GEOM_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, TOPO_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, SELECT_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, SUBSEL_TYPE_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, TEXMAP_CHAN_NUM);
	iv &= os->obj->ChannelValidity (t, VERT_COLOR_CHAN_NUM);


	if ((inkType == 7) && (os->obj->IsSubClassOf(triObjectClassID)))
		{
		TriObject *tobj = (TriObject*)os->obj;
		tobj->mesh.vertSel.ClearAll();

		tobj->mesh.selLevel = MESH_VERTEX;
		tobj->mesh.SetDispFlag(DISP_VERTTICKS | DISP_SELVERTS);
		}
	for (int i = 0; i<os->obj->NumPoints() ;i++)
		{
		Point3 p;
		p = os->obj->GetPoint(i);
		p += (DeltaList[i] + TempList[i]) * effect;
		os->obj->SetPoint(i,p);
		if ((SelectVertList[i]) && (os->obj->IsSubClassOf(triObjectClassID)))
			{
			TriObject *tobj = (TriObject*)os->obj;
			tobj->mesh.vertSel.Set(i);
			}
		}

//object type nurbs or patch
	if ( (os->obj->IsSubClassOf(triObjectClassID)) && ip)
		{
		isMesh = TRUE;
		TriObject *tobj = (TriObject*)os->obj;
		if (redoNormalList) 
			{
			BuildNormals(&tobj->mesh);
			msh = tobj->mesh;
	
			ModContextList mcList;
			INodeTab nodeList;
			ip->GetModContexts(mcList,nodeList);

			meshSearchTree.SetNormalizeMesh(FALSE);
			meshSearchTree.SetDepth(8);
			ViewExp *vpt = ip->GetActiveViewport();
			Matrix3 tm, basetm;
			basetm = nodeList[0]->GetObjectTM(t);
			tm = basetm * tm;
			vpt->GetAffineTM(tm);
			meshSearchTree.BuildTree(&msh,tm);
			ip->ReleaseViewport(vpt);
			nodeList.DisposeTemporary();
			redoNormalList = FALSE;
			}
	
		}
	else isMesh = FALSE;

	os->obj->PointsWereChanged();
	os->obj->UpdateValidity (GEOM_CHAN_NUM, iv);
	os->obj->UpdateValidity (TOPO_CHAN_NUM, iv);
	os->obj->UpdateValidity (SELECT_CHAN_NUM, iv);
	os->obj->UpdateValidity (SUBSEL_TYPE_CHAN_NUM, iv);
	os->obj->UpdateValidity (TEXMAP_CHAN_NUM, iv);
	os->obj->UpdateValidity (VERT_COLOR_CHAN_NUM, iv);

	}

Interval AFRMod::LocalValidity(TimeValue t)
	{
	Interval iv = FOREVER;
	Point3 pt;
//	pblock->GetValue(PB_FALLOFF,t,v,iv);
	iv.Set(t,t);
	return iv;
	}

Matrix3 CompMatrix(TimeValue t,INode *inode,ModContext *mc)
	{
	Interval iv;
	Matrix3 tm(1);	
	if (inode) 
		tm = tm * inode->GetObjTMBeforeWSM(t,&iv);
	return tm;
	}

void AFRMod::DrawCrossSection(Point3 a, Point3 align, float length,float strenght, GraphicsWindow *gw)

{

#define NNUM_SEGS	32

Point3 plist[NNUM_SEGS+1];
Point3 mka,mkb,mkc,mkd;

align = Normalize(align);
	{
	int ct = 0;
	float angle = TWOPI/float(NNUM_SEGS) ;
	Matrix3 rtm = RotAngleAxisMatrix(align, angle);
	Point3 p(0.0f,0.0f,0.0f);
	if (align.x == 1.0f)
		{
		p.z = length;
		}
	else if (align.y == 1.0f)
		{
		p.x = length;
		}
	else if (align.z == 1.0f)
		{
		p.y = length;
		}
	else if (align.x == -1.0f)
		{
		p.z = -length;
		}
	else if (align.y == -1.0f)
		{
		p.x = -length;
		}
	else if (align.z == -1.0f)
		{
		p.y = -length;
		}
	else 
		{
		p = Normalize(align^Point3(1.0f,0.0f,0.0f))*length;
		}

	for (int i=0; i<NNUM_SEGS; i++) {
		p = p * rtm;
		plist[ct++] = p;
		}

	p = p * rtm;
	plist[ct++] = p;


	for (i=0; i<NNUM_SEGS+1; i++) 
		{
		plist[i].x += a.x;
		plist[i].y += a.y;
		plist[i].z += a.z;
		}
	}
mka = plist[15];
mkb = plist[3];
mkc = plist[7];
mkd = plist[11];

gw->polyline(NNUM_SEGS+1, plist, NULL, NULL, 0);

plist[0] = a;
plist[1] = a + (align * strength);

gw->polyline(2, plist, NULL, NULL, 0);

}



int AFRMod::Display(
		TimeValue t, INode* inode, ViewExp *vpt, 
		int flagst, ModContext *mc)
	{
	GraphicsWindow *gw = vpt->getGW();
	Matrix3 tm = CompMatrix(t,inode,mc);
	int savedLimits;


	gw->setRndLimits((savedLimits = gw->getRndLimits()) & ~GW_ILLUM);
	gw->setTransform(tm);
	
	if (inPaint)
		{
		DrawCrossSection(paintPoint, paintNorm, eradius, strength, gw); 
		}

//	gw->marker(&pt[1],HOLLOW_BOX_MRKR);

	gw->setRndLimits(savedLimits);
	return 0;

	}





RefTargetHandle AFRMod::GetReference(int i)
	{
	switch (i) {
		case PBLOCK_REF: return (RefTargetHandle)pblock;
		default: return NULL;
		}
	}

void AFRMod::SetReference(int i, RefTargetHandle rtarg)
	{
	switch (i) {
		case PBLOCK_REF: pblock = (IParamBlock*)rtarg; break;
		}
	}



RefResult AFRMod::NotifyRefChanged(
		Interval changeInt,RefTargetHandle hTarget, 
		PartID& partID, RefMessage message)
	{
/*
	switch (message) 
		{

		}
*/

	return REF_SUCCEED;
	}



#define POINTS_CHUNK	0x0010
#define INK_TYPE_CHUNK	0x0020
#define SELECT_VERTS_CHUNK	0x0030
#define SELECT_FACES_CHUNK	0x0040

IOResult AFRMod::Load(ILoad *iload)
	{
//	Modifier::Load(iload);
	ULONG nb;
	IOResult res = IO_OK;

	iload->RegisterPostLoadCallback(
		new ParamBlockPLCB(versions,NUM_OLDVERSIONS,&curVersion,this,0));

	
	DeltaList.ZeroCount();
	TempList.ZeroCount();
	SelectVertList.ZeroCount();
	

	
	Point3 zp(0.0f,0.0f,0.0f);

	while (IO_OK==(res=iload->OpenChunk())) {
		switch (iload->CurChunkID()) {
			case POINTS_CHUNK: {
				Point3 p;
				res = iload->Read(&p,sizeof(p),&nb);
				DeltaList.Append(1,&p,1);
				TempList.Append(1,&zp,1);
				break;
				}
			case SELECT_VERTS_CHUNK: {
				BOOL b;
				res = iload->Read(&b,sizeof(b),&nb);
				SelectVertList.Append(1,&b,1);
				break;
				}

			case INK_TYPE_CHUNK: {
				int ink;
				res = iload->Read(&ink,sizeof(ink),&nb);
				inkType = ink;
				break;
				}
			}
		iload->CloseChunk();
		if (res!=IO_OK)  return res;
		}	
	
	return IO_OK;
	}

IOResult AFRMod::Save(ISave *isave)
	{

//	Modifier::Save(isave);

	ULONG nb;

	isave->BeginChunk(INK_TYPE_CHUNK);
	isave->Write(&inkType,sizeof(inkType),&nb);
	isave->EndChunk();	

		
	for (int i = 0; i< DeltaList.Count(); i++)
		{
		Point3 p;
		p = DeltaList[i];
		isave->BeginChunk(POINTS_CHUNK);
		isave->Write(&p,sizeof(p),&nb);
		isave->EndChunk();	
		}

	for (i = 0; i< SelectVertList.Count(); i++)
		{
		BOOL b;
		b = SelectVertList[i];
		isave->BeginChunk(SELECT_VERTS_CHUNK);
		isave->Write(&b,sizeof(b),&nb);
		isave->EndChunk();	
		}


	
	return IO_OK;
	}




Point3 AFRDeformer::Map(int i, Point3 p)
	{
	if (i < mod->DeltaList.Count())
		{
//ebugPrint("in deformer %d point %f %f %f offset %f %f %f\n",i,p.x,p.y,p.z,mod->DeltaList[i].x,mod->DeltaList[i].y,mod->DeltaList[i].z);
		p += (mod->DeltaList[i] + mod->TempList[i]) * mod->effect;
		}
	return p;
	}